% lab10_ex_resampling_1_all.m
%
% QUICK INTRODUCTION TO SIGNAL RESAMPLING
%
% Designing filters for signal re-sampling (up-sampling, down-sampling)
% Signal Interpolation (up-sampling) and decimation (down-sampling) filters
%
% SUMMARY
% Let assume that we have signals already sampled BUT with different frequencies:
% 8 kHz, 32 kHz, 44.1 kHz, 48 kHz (they are coming from different sources).
% But we want to change the sampling frequency digitally, i.e. resample the signal:
% - up-sample (interpolate them), in Matlab: interp(), resample(), upfirdown(),
% - down-sample (decimate them):  in Matlab: decimate(), resample(), upfirdown().
% Such operation is necessary when we want to do, for example, digital mixing of signals.
% How are we doing this?
% 
% UP-SAMPLING - K times UP
% We are appending (K-1) zeros between every two samples and pass the signal through the
% Low-Pass filter designed for cutting-frequency f0 = (fs/2)/K (interpolation filter).
% We can very easily design such filter using the window method.
%
% DOWN-SAMPLING - L times DOWN
% In this case we have to leave only every L-th sample and the rest remove.
% But doing this directly we could violate the sampling theorem telling that
% each signal sinusoid should have more than 2 samples for period.
% Therefore, we should first reduce the signal bandwidth L-times
% and then decimate the signal (take its every L-th sample).
% For this purpose we design the LowPass filter having slightly lower frequency
% than f0 = (fs/2)/L (decimation filter), e.g. (fs/2)/L-fback.
% We can very easily design such filter using the window method.
% After its usage only every L-th signal sample is left. 
%
% FRACTIONAL RE-SAMPLING
% If both filters of interpolator and decimator work in a cascade,
% offering K/L-time re-sampling, only this one with lower frequency f0 is needed.
% The interpolator should be always first in K/L resampling.
%
% EFFICIENCY - Polyphase implementation
% If interpolation/decimation ratio is very high, i.e. 1/K or 1/L is a very small number,
% it is better to use a cascade of filters with smaller up/down sampling ratio,
% i.e. instead of filter with  ratio 1/100 we could use a cascade of filters
% with ratios: 1/4, 1/5, 1/5.
% In interpolation we should avoid multiplying filter weights by inserted zeros.
% In decimation we should avoid  calculation of samples which will be removed.
% Polyphase implementations of interpolating and decimating solve these two problems.
%
% Matlab function: interp(), decimate(), resample()

clear all; close all;

fs = 1000;  dt=1/fs;   % sampling frequency, sampling period
N  = 2000;             % number of signal samples
K  = 5;                % UP-sampling order - K-times UP (first)
L  = 2;                % DOWN-sampling order - L-times DOWN (next)
 
                       % FOR DISPLAY ONLY:
R=200; m = 1:R;        % some signal indexes befere UP (interpolation)
mi = 1:K*R;            % some signal indexes after UP (interpolation)
md = 1:round(K*R/L);   % some signal indexes after DOWN (decimation)

% For up-sampling
Ni = K*N;                    % number of samples after interpolation
fsi = K*fs;                  % samplig frequency after UP-sampling (interpolation)
dti=dt/K; ti=dti*(0:Ni-1);   % sampling period, sampling time after interpolation

% For down-sampling
Nd = floor(Ni/L);            % number of samples after decimation
fsd = fsi/L;                 % samplig frequency after DOWN-sampling (decimation)
dtd=dti*L; td=dtd*(0:Nd-1);  % sampling period, sampling time after decimation

% Signal
n=0:N-1; t=dt*n;             % sampling period, sample indexes, sampling time
x = feval('sig2resample',t); % signal to be UP-sampled calculated in s2r.m
figure('Name','Signal');
plot(t(m),x(m),'rs-'); xlabel('t [s]'); title('x(t)'); grid; pause

% Matlab UP-sampling and DOWN-sampling
  if(1) % How many samples per period we have?
     figure('Name','Matlab');
     yi = interp(x,K);      plot(yi(mi),'ro-'); title('Matlab UP interp()');  pause
     yd = decimate(yi,L);   plot(yd(md),'ro-'); title('Matlab UP-DOWN interp()+ decimate()'); pause
     yr = resample(x,K,L);  plot(yr(md),'ro-'); title('Matlab UP-DOWN resample()'); pause
     plot(t(m),x(m),'rs-',ti(mi),yi(mi),'rx-',td(md),yd(md),'bo-',td(md),yr(md),'go-'); pause
  end

% ###########
% UP-SAMPLING
% ###########

% Zero appending (insertion)

if(1) % Slow
   xz=[];                              % initialization of zero-appended signal
   for n=1:N                           % N-time loop
       xz = [ xz x(n) zeros(1,K-1) ];  % copy one sample, then append K-1 zeros
   end
else  % Fast
   xz = zeros(1,Ni);                   % only K*N zeros 
   xz(1:K:Ni ) = x(1:N);               % copy signal samples into each mult*K-th positions
end    
figure('Name','Signal with zeros');
plot(t(m),x(m),'rs',ti(mi),xz(mi),'bx'); xlabel('t [s]'); title('With ZEROS x(t)'); grid; pause

% FFT spectra of signals x(n) and xz(n)
X  = abs(fft(x))/N;
Xz = abs(fft(xz))/N;
df=1/(N*dt); f=df*(0:N-1);
dfi=1/(Ni*dti); fi=dfi*(0:Ni-1);
figure('Name','FFT Spectrum of signal with zeros');
plot(fi,Xz,'bx-',f,X,'rs-'); xlabel('f [Hz]'); title('SPECTRA before and after zero insertion'); grid; pause

% Design of LowPass interpolator filter impulse response (coefficients, weights).
% Using the window method.

M = 200;                    % half-length of the filter (the whole is 2*M+1)
n = -M : M;                 % indexes of filter weights
fc = (fs/2)/fsi; % 1/(2*K)  % normaled cut-off freq; max signal frequency / new sampling
h = (2*fc)*sin(2*pi*fc*n) ./ (2*pi*fc*n);  % impulse response (weights)
h(M+1) = 2*fc;              % central weight, the (M+1)-th
h = K*h;                    % scaling of weights due zeros appened to signal

w = kaiser(2*M+1,12);             % Kaiser window for improving frequency response of the filter
h = h .* w';                      % application of the window
% h = K*fir1( 2*M,2*fc,'low',w);  % Matlab function call doing the same

figure('Name','Interpolator h(n)');
stem(n,h); grid; title('Interpolator h(n)'); pause   % display of filter weights
[ H, f ] = freqz(h,1, 1000, fsi);                    % calculation of filter frequency response
figure('Name','Interpolator H(f)');
plot(f,20*log10(abs(H)));                            % plot of filter frequency response
title('interpolator |H(f)|'); xlabel('f [Hz]'); grid; pause

% FIR filtering - xz convolved with h (weighted avering of xz using weights h) 
xi = conv(xz,h); 

% Synchronization of input and output samples - due to start-stop transient effects 
ti = ti(M+1:Ni-M);    % M first and last samples of the output are transients to be removed
xz = xz(M+1:Ni-M);    % the (M+1)-th sample of the input signal corresponds to
xi = xi(2*M+1:Ni);    % the (2*M+1)-th sample of the output signal (after the filter)
figure('Name','Signal INTERPOLATED');
plot(ti(mi),xz(mi),'rs',ti(mi),xi(mi),'bx-'); title('INTERPOLATED: RED xz(t) BLUE xint(t)'); pause

xiref = feval('sig2resample',ti);        % reference signal for UP-sampling
MaxERR_INT = max(abs(xiref-xi)), pause
figure('Name','ERROR of DOWN-samplig')
plot(ti, xiref-xi, 'b-'); xlabel('t [s]'); title('ERROR after UP-sampling'); grid; pause

% EXTRA ++++++++++++++++++++++++++++
% Fast polyphase interpolation - elimination of zero insertion by muliplication by zeros
% See lab10_ex_interp.m

Nh=2*M+1; Nz = rem(Nh,K); Napp=K-Nz; h(Nh+1:Nh+Napp) = zeros(1,Napp);  Nh=Nh+Napp;  % appending zeros at the end of h(n)
xipp = zeros(1,Ni);                             % initization of xipp signal
for k=1:K                                       % polyphase components of h(n) mult by x(n)
    xipp(k:K:Ni) = filter( h(k:K:end), 1, x );  % output are combained together 
end
xipp = xipp(2*M+1:Ni); % (2*M+1)-th sample of output signal (after filter)
MaxERR_INTpp = max(abs(xiref-xipp)), pause
figure; plot(ti(mi),xi(mi),'bx-',ti(mi),xipp(mi),'go-'); grid; xlabel('t [s]'); title('INTERPOLATED: xi(t) and xipp(t)'); pause
% ++++++++++++++++++++++++++++++++++

% #############
% DOWN-SAMPLING
% #############

% Design of LowPass decimator filter coefficients - using window method

M = 200;                           % half-length of the filter (the whole is 2*M+1)
n = -M : M;                        % indexes of filter weights 
fback = 40;                        % cut-off frequency shifted-left in respect to f3dB
fc = ((fsi/2)/L-fback) / fsi;      % fback because the stopband should start from (fsi/2)/L
h = 2*fc * sin(2*pi*fc*n) ./ (2*pi*fc*n);   % filter weights (analitically derived)
h(M+1) = 2*fc;                              % the central (M+1)-th weight

w = kaiser(2*M+1,10);              % Kaiser window for improving frequency response of the filter  
h = h .* w';                       % application of the window
% h = fir1( 2*M,2*fc,'low',w);     % Matlab function call doing the same

figure('Name','Decimator h(n)');
stem(n,h); grid; title('Decimator h(n)'); pause
[ H, f ] = freqz(h,1, 1000, fsi);  % frequency response of the filter
figure('Name','Decimator H(f)');
plot(f,20*log10(abs(H))); title('Decimator |H(f)|'); xlabel('f [Hz]'); grid; pause
axis([ 1100 1400 -160 20 ]); pause

% FIR filtering - xi is convolved with h (weighted avering of xi using weights h)
xd = conv(xi,h);

% Removal of transients in the beginning and in the end of the signal xd()
% and decimation of samples (leaving every L-th sample)
xd = xd(2*M+1 : L : end-2*M);
td = ti(M+1   : L : end-M);

% Synchronization of xi() to xd()
ti = ti(M+1:end-M);
xz = xz(M+1:end-M);
xi = xi(M+1:end-M);
figure('Name','Signal ALL')
plot(ti(mi),xz(mi),'rs',ti(mi),xi(mi),'bx-',td(md),xd(md),'ko'); title('DECIMATED: RED x(t) BLUE xint(t) BLACK xdec(t)'); grid; pause

xdref = feval('sig2resample',td);       % reference signal for DOWN-sampling
MaxERR_DEC = max(abs(xdref-xd)), pause
figure('Name','ERROR of DOWN-samplig #1 - calculation of all samples')
plot(td, xdref-xd, 'b-'); xlabel('t [s]'); title('ERROR of DOWN-sampling'); grid; pause

% Faster decimation (L times DOWN) - samples to be removed are not computed
Ni = length(xi);
Nh = 2*M+1;
Nd  = floor((Ni-Nh)/L)+1;    % signal lenght after filtering and decimation
xd2 = zeros(1,Nd);           % initization of xd2(n) signal
for k=1:Nd
    nstart = Nh + (k-1)*L;   % signal filtering with step L
    if(0)  % not-polyphase notation
       xd2(k) = sum( xi( nstart : -1 : nstart-(Nh-1) ) .* h );
    else   % polyphase notation   
       xd2(k) = 0;
       for m=1:L
           xd2(k) = xd2(k) + sum( xi( nstart-(m-1) : -L : nstart-(Nh-1) ) .* h(m:L:end) );
       end
    end   
end
td2 = ti(M+1:L:end-M);
xdref = feval('sig2resample',td2);       % reference signal for DOWN-sampling
MaxERR_DEC2 = max(abs(xdref-xd2)), pause
figure('Name','ERROR of DOWN-samplig #2 - calculation of necessary samples');
plot(td(md),xd(md),'r.-',td2(md),xd2(md),'b.-'); grid; xlabel('t [s]'); title('xd1(t) and xd2(t)'); pause

% EXTRA ++++++++++++++++++++++++++++
% Fast polyphase decimation (L times DOWN) - samples to be removed are not computed
% See lab10_ex_decim.m

Nz = rem(Ni,L); Napp=L-Nz; xi(Ni+1:Ni+Napp) = zeros(1,Napp);  Ni=Ni+Napp;  % appending zeros at the end of xi(n)
Nz = rem(Nh,L); Napp=L-Nz;  h(Nh+1:Nh+Napp) = zeros(1,Napp);  Nh=Nh+Napp;  % appending zeros at the end of h(n)

xdref = filter(h,1,xi); xdref = xdref(1:L:end);

xi = [ zeros(1,L-1) xi(1:Ni-(L-1)) ];
xdpp = zeros(1,Ni/L);                   % initization of xDOWN signal
for m=1:L
    xdpp = xdpp + filter( h(m:L:Nh), 1, xi(L-m+1:L:Ni) );
end
plot(md,xdref(md),'ro-',md,xdpp(md),'bx-'); title('TEST'); pause

xdpp = xdpp((2*M)/L+1:Nd);
tdpp = td2(1:end-(2*M)/L);
xdref = feval('sig2resample',tdpp);       % reference signal for DOWN-sampling
MaxERR_DECpp = max(abs(xdref-xdpp)), pause
figure('Name','Polyphase decimation vs. reference signal');
plot(tdpp(md),xdref(md),'r.-',tdpp(md),xdpp(md),'b.-'); grid; xlabel('t [s]'); title('xdref(t) and xdpp(t)'); pause
figure('Name','Polyphase decimation - two versions');
plot(td2(md),xd2(md),'r.-',td2(md),xdpp(md),'b.-'); grid; xlabel('t [s]'); title('xd2(t) and xdpp(t)'); pause
% ++++++++++++++++++++++++++++++++++
